# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Till Huckeman

# flake8: noqa: F405

# %% Experiment Setup

import threading

import numpy as np
import yaml
from qcodes.dataset import (
    Measurement,
    experiments,
    initialise_or_create_database_at,
    load_by_run_spec,
    load_or_create_experiment,
)
from qcodes.station import Station

from qumada.instrument.buffered_instruments import BufferedDummyDMM as DummyDmm
from qumada.instrument.buffers.buffer import (
    load_trigger_mapping,
    map_triggers,
    save_trigger_mapping,
)
from qumada.instrument.custom_drivers.Dummies.dummy_dac import DummyDac
from qumada.instrument.mapping import (
    DUMMY_DMM_MAPPING,
    add_mapping_to_instrument,
    map_terminals_gui,
)
from qumada.instrument.mapping.Dummies.DummyDac import DummyDacMapping
from qumada.measurement.device_object import *
from qumada.measurement.scripts import (
    Generic_1D_parallel_Sweep,
    Generic_1D_Sweep,
    Generic_1D_Sweep_buffered,
    Generic_nD_Sweep,
    Timetrace,
    Timetrace_with_Sweeps_buffered,
)
from qumada.utils.generate_sweeps import generate_sweep
from qumada.utils.load_from_sqlite_db import load_db

# %% Only required to simulate buffered instruments
# As we have only dummy instruments that are not connected, we have to use a global
# trigger event for triggering.
trigger = threading.Event()

# Setup qcodes station and load instruments
station = Station()

# The dummy instruments have a trigger_event attribute as replacement for
# the trigger inputs of real instruments.

# Creating QCoDeS Instrument. For buffered instruments we have to use the QuMada buffered version.
dmm = DummyDmm("dmm", trigger_event=trigger)
# Adding the QuMada mapping
add_mapping_to_instrument(dmm, mapping=DUMMY_DMM_MAPPING)
# Add QCoDes Instrument to station
station.add_component(dmm)

dac1 = DummyDac("dac1", trigger_event=trigger)
add_mapping_to_instrument(dac1, mapping=DummyDacMapping())
station.add_component(dac1)

dac2 = DummyDac("dac2", trigger_event=trigger)
add_mapping_to_instrument(dac2, mapping=DummyDacMapping())
station.add_component(dac2)
# %% Load database for data storage. This will open a window.
# Alternatively, you can use initialise_or_create_database_at from QCoDeS
load_db()
# We need to create an experiment in the QCoDeS database
load_or_create_experiment("test", "dummy_sample")
# %% Setup buffers (only need for buffered measurements).
# Those buffer settings specify how the triggers are setup and how the data is recorded.
buffer_settings = {
    # We don't have to specify threshold and mode for our dummy instruments
    # "trigger_threshold": 0.005,
    # "trigger_mode": "digital",
    "sampling_rate": 20,
    "num_points": 100,
    "delay": 0,
}

# This tells a measurement script how to start a buffered measurement.
# "Hardware" means that you want to use a hardware trigger. To start a measurement,
# the method provided as "trigger_start" is called. The "trigger_reset" method is called
# at the end of each buffered line, in our case resetting the trigger flag.
# For real instruments, you might have to define a method that sets the output of your instrument
# to a desired value as "trigger_start". For details on other ways to setup your triggers,
# check the documentation.

buffer_script_settings = {
    "trigger_type": "hardware",
    "trigger_start": trigger.set,
    "trigger_reset": trigger.clear,
}

# %% Parameters
# This dictionary defines your device. "ohmic", "gate1" and "gate2" are terminals, representing parts of your
# device.
# Each terminal can have multiple parameters that represent values of the terminals
# (here "current" and "voltage"). A real Ohmic connected to a lockin-amplifier could for example have
# additional parameters such as "amplitude" or "frequency". As our dummy_dmm doesn't have those parameters,
# we cannot use them here. Each parameter has to be mapped to a parameter of a QCoDeS instrument later.

parameters = {
    "ohmic": {
        "current": {"type": "gettable"},
    },
    "gate1": {"voltage": {"type": "static"}},
    "gate2": {"voltage": {"type": "static"}},
}

# %% Device
# This creates a device from the dictionary we just specfied.
# device.mapping() will open a GUI that allows you to map the parameters of the device to parameters
# of the measurement instruments (in this case the dummy instruments).
# You can simply use the automated mapping in this case (for details check the documentation).
# Then next two lines save the buffer and trigger settings to the device so we don't have to specify them
# again when we want to run measurements.
# map_triggers will ask you to specify which trigger input to use. Use "software" here.
# You can also load and save parameter
# and trigger mappings to from/to files (see documentation).

device = QumadaDevice.create_from_dict(parameters, station=station, namespace=globals())
device.mapping()
device.buffer_script_setup = buffer_script_settings
device.buffer_settings = buffer_settings
map_triggers(station.components)
# %% Play around

# Change parameter values by calling them
gate2.voltage(0.2)
print(gate2.voltage())

# In case you want to change a voltage parameter, you can also call the terminal directly.
# This works only for getting and setting voltages!
gate1(-0.3)

print(ohmic.current())

# This prints all voltages of the device
device.voltages()

# %% Run simple measurements
# This will ramp gate1 from its current voltage to 0.4 V and record all parameters of type "gettable".
# In this case, this is ohmic.current.

print(ohmic.current.type)
print(gate2.voltage.type)

gate1.voltage.measured_ramp(0.4, buffered=True)

# %% This will also record gate2.voltage. As the dummy dac has no buffer, we cannot record
# a buffered measurement if we want to measure the voltage. The "start" argument can be used to
# start from a different position than the current voltage of gate1. QuMada will by default ramp
# gate1.voltage to the starting point before starting the measurement to avoid artifacts from
# a voltage jump prior to the measurement.
# QuMada measurement scripts will automatically name your measurement unless you explicitely provide
# a name.

gate2.voltage.type = "gettable"
gate1.voltage.measured_ramp(0.4, start=-0.3, name="mymeasurement", buffered=False)  # noqa: F405

gate2.voltage.type = ""  # Don't record it in future measurements.

# %% Save setpoints

print("State 1:")
device.voltages()  # Current state

device.save_state("state1")  # Save current configuration

gate2(0.19)

print("State 2:")
device.voltages()  # New state

device.set_state("state1")  # Reset to stored state
print("Reset to State 1:")
device.voltages()

# %% 2D Scans
# 2D Scans are centered around the current voltage setpoints and return to the
# setpoints are they are done.
# This one will sweep gate1 from 0, to 0.2 V and gate2 from -0.15 to 0.15 V


gate1(0.1)
gate2(0)

print("Before measurement:")
device.voltages()

device.sweep_2D(gate1.voltage, gate2.voltage, 0.2, 0.3, buffered=True)

print("After measurement:")
device.voltages()

# %% Timetraces are recorded at the current setpoint. All measurements return a list of QCoDeS datasets.

data = device.timetrace(duration=10, timestep=0.3, buffered=True)

# If you provide measurement settings such as the timestep or num_points that do not
# match the buffer settings, the buffer settings will be (temporarily) altered to.
# %%
# device.save_to_dict() returns a parameter dictionary based on the current state of the device.

# Alter parameters for measurement...
gate1.voltage.setpoints = np.linspace(0, 0.3, 100)
gate2.voltage.setpoints = np.linspace(0, 0.1, 100)
gate1.voltage.type = "dynamic"
gate2.voltage.type = "dynamic"

print(device.save_to_dict())

# You can feed it back into any measurement script as parameters:
script = Generic_1D_parallel_Sweep()
script.setup(
    device.save_to_dict(),
    metadata=None,
)

# We can provide the mapping from the device object as argument to skip the mapping
map_terminals_gui(station.components, script.gate_parameters, device.instrument_parameters)

script.run()


# %%
# A more convenient way to run arbitrary measurement scripts, is to use
# device.run_measurement, as this uses the advantages of the device.
# Lets for example run a timetrace with sweeps

device.run_measurement(
    script=Timetrace_with_Sweeps_buffered,
    dynamic_params=[gate1.voltage],
    setpoints=[np.linspace(-0.1, 0.1, 100)],
    static_params=[gate2.voltage],
    static_values=[0.3],
    buffered=True,
    name="Timetrace with sweeps",
    duration=20,
)


# The duration is a keyword argument that is passed on to the used measurement
# script.
# It is possible to only provide the script, in this case current settings
# of the device object are used.



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Bertha Lab Setup
# - Daniel Grothe
# - Till Huckeman

import logging

from qcodes.parameters import Parameter

from qumada.instrument.custom_drivers.QDevil.QDAC2 import QDac2
from qumada.instrument.mapping import QDAC2_MAPPING
from qumada.instrument.mapping.base import InstrumentMapping

logger = logging.getLogger(__name__)


class QDac2Mapping(InstrumentMapping):
    def __init__(self):
        super().__init__(QDAC2_MAPPING)

    def ramp(
        self,
        parameters: list[Parameter],
        *,
        start_values: list[float] | None = None,
        end_values: list[float],
        ramp_time: float,
        sync_trigger=None,
        **kwargs,
    ) -> None:
        """
        Qumada ramp method using the QDacII dc_sweep to ramp smoothly.
            parameters: List of parameters to be set.
                        Elements have to be from the same QDACII
            start_values:  List of start values for the parameters (optional).
                        Current value will be used if None.
            stop_values: List of end values for the parameters.
            delay:      Time in between to setpoints in s. Has to be >=1e-6 s
            sync_trigger [int|None]: Ext Trigger output (1-6) to use for sync triggering
                        Sends a trigger pulse when the ramp is started.
        kwargs:
            num_points [int]: Number of datapoints to use, defines smoothness. Default
                        is 1000/sec.
            trigger_width [float]: Length of sync trigger pulse in s. Default 1e-3.
            trigger_polarity [str]: Direction of sync trigger pulse.

        """

        trigger_width = kwargs.get("trigger_width", 1e-3)
        trigger_polarity = kwargs.get("trigger_polarity", "norm")
        points = kwargs.get("num_points", int(ramp_time * 1000))
        wait_time = ramp_time / points
        if wait_time < 1e-6:
            raise Exception("QDac 2 wait_time is to small (<1e-6s)")

        assert len(parameters) == len(end_values)
        for parameter in parameters:
            assert parameter._short_name == "dc_constant_V"

        if start_values is not None:
            assert len(parameters) == len(start_values)

        if len(parameters) > 8:
            raise Exception("Maximum length of rampable parameters is 8.")

        # check, if all parameters are from the same instrument
        instruments = {parameter.root_instrument for parameter in parameters}
        if len(instruments) > 1:
            raise Exception("Parameters are from more than one instrument. This would lead to non synchronized ramps.")

        qdac: QDac2 = instruments.pop()
        assert isinstance(qdac, QDac2)

        if not start_values:
            start_values = [param() for param in parameters]

        channels = [param._instrument for param in parameters]
        for channel, start, stop in zip(channels, start_values, end_values):
            dc_sweep = channel.dc_sweep(
                start_V=start, stop_V=stop, points=points, dwell_s=wait_time, backwards=(start > stop)
            )
            # There appears to be a bug with the built in dc_sweep of the QDAC!
            # The sweep direction is always from the lower to the large value if
            # backwards==False and the other way round if it is true.
            # Once this is fixed, remove the backwards argument from function call!

        if sync_trigger is not None:
            if sync_trigger in range(1, 6):
                trigger = dc_sweep.start_marker()
                qdac.external_triggers[sync_trigger - 1].width_s(trigger_width)
                qdac.external_triggers[sync_trigger - 1].polarity(trigger_polarity)
                qdac.external_triggers[sync_trigger - 1].source_from_trigger(trigger)
            else:
                logger.warning(
                    f"{sync_trigger} is no valid sync trigger for QDac II. Choose an integer between 1 and 5!"
                )

        qdac.start_all()
        qdac.free_all_triggers()

    def pulse(
        self,
        parameters: list[Parameter],
        *,
        setpoints: list[list[float]],
        delay: float,
        sync_trigger=None,
        **kwargs,
    ) -> None:
        """
        Qumada Pulse method using the QDacII dc_list to set arbitrary pulses.
            parameters: List of parameters to be set.
                        Elements have to be from the same QDACII
            setpoints:  List of setpoint arrays.
            delay:      Time in between to setpoints in s. Has to be >=1e-6 s
            sync_trigger [int|None]: Ext Trigger output (1-6) to use for sync triggering
                        Sends a trigger pulse when the ramp is started.
        kwargs:
            trigger_width [float]: Length of sync trigger pulse in s. Default 1e-3.
            trigger_polarity [str]: Direction of sync trigger pulse.

        """

        trigger_width = kwargs.get("trigger_width", 1e-3)
        trigger_polarity = kwargs.get("trigger_polarity", "norm")
        # Make sure everything is in order...
        assert len(parameters) == len(setpoints)
        for points in setpoints:
            assert len(points) == len(setpoints[0])
            for parameter in parameters:
                assert parameter._short_name == "dc_constant_V"
        instruments = {parameter.root_instrument for parameter in parameters}
        if len(instruments) > 1:
            raise Exception(
                "Parameters are from more than one instrument. \
                This would lead to non synchronized ramps."
            )
        qdac: QDac2 = instruments.pop()
        assert isinstance(qdac, QDac2)
        if delay < 1e-6:
            raise Exception("Delay for QDacII pulse is to small (<1 us)")
        channels = [param._instrument for param in parameters]
        self.dc_lists = [
            channel.dc_list(
                voltages=points,
                dwell_s=delay,
            )
            for channel, points in zip(channels, setpoints)
        ]

        if sync_trigger is not None:
            if sync_trigger in range(1, 6):
                trigger = self.dc_lists[0].start_marker()
                qdac.external_triggers[sync_trigger - 1].width_s(trigger_width)
                qdac.external_triggers[sync_trigger - 1].polarity(trigger_polarity)
                qdac.external_triggers[sync_trigger - 1].source_from_trigger(trigger)
            else:
                logger.warning(
                    f"{sync_trigger} is no valid sync trigger for QDac II. Choose an integer between 1 and 5!"
                )
        qdac.start_all()
        qdac.free_all_triggers()

    def setup_trigger_in():
        raise Exception(
            "QDac2 does not have a trigger input \
            not yet supported!"
        )

    def clean_generators(self):
        for dc_list in self.dc_lists:
            dc_list.abort()
        self.dc_lists = []

    @staticmethod
    def query_instrument(parameters: list[Parameter]):
        """Check if all parameters are from the same instrument"""
        instruments = {parameter.root_instrument for parameter in parameters}
        if len(instruments) > 1:
            raise Exception(
                "Parameters are from more than one instrument. \
                This would lead to non synchronized ramps."
            )
        qdac: QDac2 = instruments.pop()
        assert isinstance(qdac, QDac2)
        return qdac



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Daniel Grothe
# - Till Huckeman

from qcodes.instrument_drivers.stanford_research.SR830 import SR830

from qumada.instrument.buffers import DummyDMMBuffer, MFLIBuffer, SR830Buffer
from qumada.instrument.custom_drivers.Dummies.dummy_dmm import DummyDmm
from qumada.instrument.custom_drivers.ZI.MFLI import MFLI, Session


class BufferedMFLI(MFLI):
    """Buffered version of Zurich Instruments MFLI."""

    def __init__(
        self, name: str, device: str, serverhost: str = "localhost", existing_session: Session = None, **kwargs
    ):
        super().__init__(name=name, device=device, serverhost=serverhost, existing_session=existing_session, **kwargs)
        self._qumada_buffer = MFLIBuffer(self)


class BufferedSR830(SR830):
    """Buffered version of Stanford SR830."""

    def __init__(self, name: str, address: str, **kwargs):
        super().__init__(name=name, address=address, **kwargs)
        self._qumada_buffer = SR830Buffer(self)


class BufferedDummyDMM(DummyDmm):
    def __init__(self, name: str, **kwargs):
        super().__init__(name=name, **kwargs)
        self._qumada_buffer = DummyDMMBuffer(self)



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Daniel Grothe


from __future__ import annotations

import importlib
import pkgutil
from contextlib import suppress
from types import ModuleType


def import_submodules(package: str | ModuleType, recursive: bool = True) -> dict[str, ModuleType]:
    """Import all submodules of a module, recursively, including subpackages"""
    if isinstance(package, str):
        package = importlib.import_module(package)
    results = {}
    for loader, name, is_pkg in pkgutil.walk_packages(package.__path__):
        full_name = package.__name__ + "." + name
        with suppress(ValueError, ImportError):
            results[full_name] = importlib.import_module(full_name)
        if recursive and is_pkg:
            results.update(import_submodules(full_name))
    return results



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Daniel Grothe
# - Till Huckeman


# most of the drivers only need a couple of these... moved all up here for clarity below
from __future__ import annotations

import threading
from time import sleep

import numpy as np
from qcodes.instrument import ChannelList, Instrument, InstrumentChannel
from qcodes.validators import validators as vals


# %%
class DummyDac_Channel(InstrumentChannel):
    def __init__(self, parent, name, channel):
        super().__init__(parent, name)
        self._channel = channel

        self.add_parameter("voltage", unit="V", set_cmd=None, vals=vals.Numbers(-10, 10))
        self.voltage.set(0)

    def ramp(self, start, stop, duration, num_points):
        self.parent.ramp(self, start, stop, duration, num_points)


class DummyDac(Instrument):
    def __init__(self, name, trigger_event=threading.Event(), **kwargs):
        super().__init__(name, **kwargs)
        channels = ChannelList(self, "Instrument_Channels", DummyDac_Channel)
        self._is_triggered = trigger_event

        for i in range(1, 5):
            channel = DummyDac_Channel(self, f"ch{i:02}", i)
            channels.append(channel)
            self.add_submodule(f"ch{i:02}", channel)
        # self.add_parameter("voltage", unit="V", set_cmd=None, vals=vals.Numbers(-10, 10))
        # self.voltage.set(0)
        self.add_submodule("channels", channels.to_channel_tuple())
        self.add_function("force_trigger", call_cmd=self._is_triggered.set)

    def _run_ramp(self, channel, start, stop, duration, num_points):
        for setpoint in np.linspace(start, stop, int(num_points)):
            channel.voltage(setpoint)
            sleep(duration / num_points)

    def ramp(self, channel, start, stop, duration, num_points):
        self.thread = threading.Thread(
            target=self._run_ramp,
            args=(channel, start, stop, duration, num_points),
            daemon=True,
        )
        self.thread.start()

    def ramp_channels(self, channels: list, start_values: list, stop_values: list, duration, num_points):
        self.thread = threading.Thread(
            target=self._run_ramp_channels,
            args=(channels, start_values, stop_values, duration, num_points),
            daemon=True,
        )
        self.thread.start()

    def _run_ramp_channels(self, channels: list, start_values: list, stop_values: list, duration, num_points):
        setpoints = []
        for ch, start, stop in zip(channels, start_values, stop_values):
            setpoints.append(np.linspace(start, stop, num_points))
        setpoints_inv = []
        for i in range(num_points):
            setpoints_inv.append([setpoints[j][i] for j in range(len(channels))])
        for setpoint in setpoints_inv:
            for i in range(len(channels)):
                channels[i].voltage(setpoint[i])
            sleep(duration / num_points)

    def _run_triggered_ramp(self, channel, start, stop, duration, stepsize=0.01):
        _ = self._is_triggered.wait()
        num_points = int((stop - start) / stepsize)
        for setpoint in np.linspace(start, stop, num_points):
            channel.voltage(setpoint)
            sleep(duration / num_points)

    def _run_triggered_ramp_channels(self, channels, start_values, stop_values, duration, num_points):
        setpoints = []
        for start, stop in zip(start_values, stop_values):
            setpoints.append(np.linspace(start, stop, int(num_points)))
        setpoints_inv = []
        for i in range(int(num_points)):
            setpoints_inv.append([setpoints[j][i] for j in range(len(channels))])
        _ = self._is_triggered.wait()
        for setpoint in setpoints_inv:
            for i in range(len(channels)):
                channels[i].voltage(setpoint[i])
            sleep(duration / num_points)

    def _run_triggered_pulse_channels(self, channels, setpoints, duration):
        setpoints_inv = []
        num_points = len(setpoints[0])
        for i in range(int(len(setpoints[0]))):
            setpoints_inv.append([setpoints[j][i] for j in range(len(channels))])
        _ = self._is_triggered.wait()
        for setpoint in setpoints_inv:
            for i in range(len(channels)):
                channels[i].voltage(setpoint[i])
            sleep(duration / num_points)

    def _triggered_ramp(self, channel, start, stop, duration, num_points):
        self.thread = threading.Thread(
            target=self._run_triggered_ramp,
            args=(channel, start, stop, duration, num_points),
            daemon=True,
        )
        self.thread.start()

    def _triggered_ramp_channels(self, channels, start_values, stop_values, duration, num_points):
        self.thread = threading.Thread(
            target=self._run_triggered_ramp_channels,
            args=(channels, start_values, stop_values, duration, num_points),
            daemon=True,
        )
        self.thread.start()

    def _triggered_pulse_channels(self, channels, setpoints, duration):
        self.thread = threading.Thread(
            target=self._run_triggered_pulse_channels,
            args=(channels, setpoints, duration),
            daemon=True,
        )
        self.thread.start()


# %%



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Daniel Grothe
# - Till Huckeman

from __future__ import annotations

import logging

import numpy as np
from jsonschema import validate
from qcodes.parameters import Parameter

from qumada.instrument.buffers.buffer import Buffer, BufferException
from qumada.instrument.custom_drivers.ZI.MFLI import MFLI

logger = logging.getLogger(__name__)


class MFLIBuffer(Buffer):
    """Buffer for ZurichInstruments MFLI"""

    AVAILABLE_TRIGGERS: list[str] = [
        "trigger_in_1",
        "trigger_in_2",
        "aux_in_1",
        "aux_in_2",
    ]

    TRIGGER_MODE_MAPPING: dict = {
        "continuous": 0,
        "edge": 1,
        "pulse": 3,
        "tracking_edge": 4,
        "tracking_pulse": 7,
        "digital": 6,
    }

    TRIGGER_MODE_POLARITY_MAPPING: dict = {"positive": 1, "negative": 2, "both": 3}

    GRID_INTERPOLATION_MAPPING: dict = {"nearest": 1, "linear": 2, "exact": 4}
    # TODO: What happens in exact mode? Look up limitations...

    def __init__(self, mfli: MFLI):
        self._session = mfli.session
        self._device = mfli.instr
        self._daq = self._session.modules.daq
        self._sample_nodes: list = []
        self._subscribed_parameters: list[Parameter] = []
        self._trigger: str | None = None
        self._channel = 0
        self._num_points: int | None = None
        self._num_bursts: int = 1
        self._burst_duration: float | None = None

    def setup_buffer(self, settings: dict) -> None:
        # validate settings
        validate(settings, self.settings_schema)
        self.settings: dict = settings
        device = self._device
        self._daq.device(device)

        if "channel" in settings:
            self._channel = settings["channel"]

        device.demods[self._channel].enable(True)

        # TODO: Validate Trigger mode, edge and interpolation!:
        self._daq.type(self.TRIGGER_MODE_MAPPING[settings.get("trigger_mode", "edge")])
        self._daq.edge(self.TRIGGER_MODE_POLARITY_MAPPING[settings.get("trigger_mode_polarity", "positive")])
        self._daq.grid.mode(self.GRID_INTERPOLATION_MAPPING[settings.get("grid_interpolation", "linear")])
        self.trigger = self.trigger  # Don't delete me, I am important!
        if "trigger_threshold" in settings:
            # TODO: better way to distinguish, which trigger level to set
            self._daq.level(settings["trigger_threshold"])
            self._device.triggers.in_[0].level(settings["trigger_threshold"])
            self._device.triggers.in_[1].level(settings["trigger_threshold"])
        else:
            logger.warning("No trigger threshold specified!")
        self._set_num_points()
        self._daq.delay = settings.get("delay", 0)

    @property
    def trigger(self):
        return self._trigger

    @trigger.setter
    def trigger(self, trigger: str | None) -> None:
        # TODO: Inform user about automatic changes of settings
        # TODO: This is done BEFORE the setup_buffer, so changes to trigger type will be overriden anyway?
        # print(f"Running trigger setter with: {trigger}")
        if trigger is None:
            logger.info("No Trigger provided! Setting trigger to continuous.")
            self._daq.type(0)
        elif trigger in self.AVAILABLE_TRIGGERS:
            samplenode = self._device.demods[self._channel].sample
            if trigger == "trigger_in_1":
                self._daq.triggernode(samplenode.TrigIn1)
                self._daq.type(6)
            elif trigger == "trigger_in_2":
                self._daq.triggernode(samplenode.TrigIn2)
                self._daq.type(6)
            elif trigger == "aux_in_1":
                self._daq.triggernode(samplenode.AuxIn0)
                if self._daq.type() not in (1, 3, 4, 7):
                    self._daq.type(1)
            elif trigger == "aux_in_2":
                self._daq.triggernode(samplenode.AuxIn1)
                if self._daq.type() not in (1, 3, 4, 7):
                    self._daq.type(1)
        else:
            raise BufferException(f"Trigger input '{trigger}' is not supported.")
        self._trigger = trigger

    def force_trigger(self) -> None:
        self._daq.forcetrigger(1)

    @property
    def num_points(self) -> int | None:
        return self._num_points

    @num_points.setter
    def num_points(self, num_points) -> None:
        if num_points > 8_388_608:
            raise BufferException(
                "Buffer is to small for this measurement. \
                                  Please reduce the number of data points"
            )
        self._num_points = int(num_points)

    # TODO: Define setter for other settings (e.g. burst_duration, num_bursts etc)

    def _set_num_points(self) -> None:
        """
        Calculates all required settings for the MFLI and sets the values
        accordingly
        ------
        Exception
           Exception if number of points or number of burst is overdefined.

        Returns
        -------
        None
        """
        # TODO: Include ._daq.repetitions (averages over multiple bursts)

        if all(k in self.settings for k in ("sampling_rate", "burst_duration", "num_points")):
            raise BufferException("You cannot define sampling_rate, burst_duration and num_points at the same time")

        if all(k in self.settings for k in ("sampling_rate", "duration", "num_bursts", "num_points")):
            raise BufferException(
                "You cannot define sampling rate, duration and num_burst and num_points at the same time"
            )

        if all(k in self.settings for k in ("num_bursts", "duration", "burst_duration")):
            raise BufferException("You cannnot define duration, burst_duration and num_bursts at the same time")

        if "burst_duration" in self.settings:
            self._burst_duration = self.settings["burst_duration"]

        if "duration" in self.settings:
            if "burst_duration" in self.settings:
                self._num_bursts = np.ceil(self.settings["duration"] / self._burst_duration)
            elif "num_bursts" in self.settings:
                self._num_bursts = int(self.settings["num_bursts"])
                self._burst_duration = self.settings["duration"] / self._num_bursts
            else:
                logger.info(
                    "You have specified neither burst_duration nor num_bursts. \
                      Using duration as burst_duration!"
                )
                self._burst_duration = self.settings["duration"]

        if "num_points" in self.settings:
            self.num_points = int(self.settings["num_points"])
            if "sampling_rate" in self.settings:
                self._burst_duration = float(self.num_points / self.settings["sampling_rate"])

        elif "sampling_rate" in self.settings:
            self._sampling_rate = float(self.settings["sampling_rate"])
            if self._burst_duration is not None:
                self.num_points = int(np.ceil(self._sampling_rate * self._burst_duration))
            elif all(k in self.settings for k in ("duration", "num_bursts")):
                self._burst_duration = float(self.settings["duration"] / self.settings["num_bursts"])

        self._daq.count(self._num_bursts)
        self._daq.duration(self._burst_duration)
        self._daq.grid.cols(self.num_points)

    def read(self) -> dict:
        data = self.read_raw()
        result_dict = {}
        for parameter in self._subscribed_parameters:
            node = self._get_node_from_parameter(parameter)
            key = next(key for key in data.keys() if str(key) == str(node))
            result_dict[parameter.name] = data[key][0].value
            if "timestamps" not in result_dict:
                result_dict["timestamps"] = data[key][0].time
        return result_dict

    def read_raw(self) -> dict:
        return self._daq.read()

    def subscribe(self, parameters: list[Parameter]) -> None:
        for parameter in parameters:
            node = self._get_node_from_parameter(parameter)
            if node not in self._sample_nodes:
                self._subscribed_parameters.append(parameter)
                self._sample_nodes.append(node)
                self._daq.subscribe(node)

    def unsubscribe(self, parameters: list[Parameter]) -> None:
        for parameter in parameters.copy():
            node = self._get_node_from_parameter(parameter)
            if node in self._sample_nodes:
                self._sample_nodes.remove(node)
                self._subscribed_parameters.remove(parameter)
                self._daq.unsubscribe(node)

    def is_subscribed(self, parameter: Parameter) -> bool:
        return parameter in self._subscribed_parameters

    def start(self) -> None:
        self._daq.execute()

    def stop(self) -> None:
        self._daq.raw_module.finish()

    def is_ready(self) -> bool: ...

    def is_finished(self) -> bool:
        return self._daq.raw_module.finished()

    def _get_node_from_parameter(self, parameter: Parameter):
        return self._device.demods[self._channel].sample.__getattr__(parameter.signal_name[1])



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Daniel Grothe
# - Till Huckeman

# Ignore flake8 and mypy, as these file is going to be redone either way.
# TODO: Remove these comments then
# flake8: noqa
# type: ignore

# %% Experiment Setup
# As we have only dummy instruments that are not connected, we have to use a global
# trigger event for triggering.
import threading

import numpy as np
import yaml
from qcodes.dataset import (
    Measurement,
    experiments,
    initialise_or_create_database_at,
    load_by_run_spec,
    load_or_create_experiment,
)
from qcodes.station import Station

from qumada.instrument.buffered_instruments import BufferedDummyDMM as DummyDmm
from qumada.instrument.buffers.buffer import (
    load_trigger_mapping,
    map_triggers,
    save_trigger_mapping,
)
from qumada.instrument.custom_drivers.Dummies.dummy_dac import DummyDac
from qumada.instrument.mapping import (
    DUMMY_DMM_MAPPING,
    add_mapping_to_instrument,
    map_terminals_gui,
)
from qumada.instrument.mapping.Dummies.DummyDac import DummyDacMapping
from qumada.measurement.scripts import (
    Generic_1D_parallel_asymm_Sweep,
    Generic_1D_parallel_Sweep,
    Generic_1D_Sweep,
    Generic_1D_Sweep_buffered,
    Generic_nD_Sweep,
    Timetrace,
)
from qumada.utils.generate_sweeps import generate_sweep, replace_parameter_settings
from qumada.utils.GUI import open_web_gui
from qumada.utils.load_from_sqlite_db import load_db
from qumada.utils.ramp_parameter import *

# %%
trigger = threading.Event()

# Setup qcodes station
station = Station()

# The dummy instruments have a trigger_event attribute as replacement for
# the trigger inputs of real instruments.

dmm = DummyDmm("dmm", trigger_event=trigger)
add_mapping_to_instrument(dmm, mapping=DUMMY_DMM_MAPPING)
station.add_component(dmm)

dac = DummyDac("dac", trigger_event=trigger)
add_mapping_to_instrument(dac, mapping=DummyDacMapping())
station.add_component(dac)

dac2 = DummyDac("dac2", trigger_event=trigger)
add_mapping_to_instrument(dac2, mapping=DummyDacMapping())
station.add_component(dac2)
# %% Load database for data storage
load_db()
# %% Setup measurement
buffer_settings = {
    # "trigger_threshold": 0.005,
    # "trigger_mode": "digital",
    "sampling_rate": 20,
    "duration": 5,
    "burst_duration": 5,
    "delay": 0,
}

# %% Measurement Setup
parameters = {
    "ohmic": {
        "voltage": {"type": "gettable"},
        "current": {"type": "gettable"},
    },
    "gate1": {"voltage": {"type": "dynamic", "setpoints": np.linspace(0, np.pi, 100), "value": 0}},
    "gate2": {"voltage": {"type": "dynamic", "setpoints": np.linspace(0, np.pi, 100), "value": 0}},
}
# %%
script = Generic_1D_Sweep_buffered()
script.setup(
    parameters,
    metadata=None,
    buffer_settings=buffer_settings,
    trigger_type="hardware",
    trigger_start=trigger.set,
    trigger_reset=trigger.clear,
)

map_terminals_gui(station.components, script.gate_parameters)
map_triggers(station.components)

# %% Run measurement
script.run()



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Till Huckeman

import logging

import numpy as np
from qcodes.dataset.measurements import Measurement
from qcodes.parameters.parameter import Parameter
from qcodes.parameters.specialized_parameters import ElapsedTimeParameter
from qutil.measurement.spectrometer import Spectrometer, daq

from qumada.measurement.measurement import MeasurementScript
from qumada.utils.utils import naming_helper

logger = logging.getLogger(__name__)


class Measure_Spectrum(MeasurementScript):
    """
    kwargs:
        store_timetrace: Bool|True. Stores timetrace as QCoDeS measurement
        store_spectrum: Bool|True:. Stores spectrum as QCoDes measurement
    """

    def run(self) -> list:
        self.initialize()
        naming_helper(self, default_name="Spectrum")
        settings = self.settings
        store_timetrace = settings.get("store_timetrace", True)
        store_spectrum = settings.get("store_spectrum", True)
        # module = settings.get("module", "scope")

        # TODO: Check if instrument is supported
        # TODO: Cases for different instruments/modes

        if len(self.gettable_channels) != 1:
            if self.priorities == {}:
                raise AssertionError("For this type of measurement exactly 1 gettable channel is required!")
            else:
                candidates = self.priorities[min(self.priorities.keys())]["channels"]
                intersect = [ch for ch in self.gettable_channels if ch in candidates]
                assert len(intersect) == 1
                self.dependent_param = intersect[0]
                instrument = self.dependent_param.root_instrument
        else:
            self.dependent_param = self.gettable_channels[0]
            instrument = self.dependent_param.root_instrument

        if settings.get("module", "scope") == "scope":
            setup, acquire = daq.zhinst.MFLI_scope(instrument.instr.session, instrument.instr)
        else:
            setup, acquire = daq.zhinst.MFLI_daq(instrument.instr.session, instrument.instr)
        try:
            self.spectrometer = settings.pop("spectrometer")
        except KeyError:
            self.spectrometer = Spectrometer(setup, acquire)
        self.spectrometer.take(self.measurement_name, **settings)
        results = self.spectrometer[-1]
        if store_timetrace:
            self._save_data_to_db(results=results, data_type="timetrace")
        if store_spectrum:
            self._save_data_to_db(results=results, data_type="spectrum")
        self.settings["spectrometer"] = self.spectrometer
        return self.spectrometer

    def _save_data_to_db(
        self,
        results: dict,
        data_type: str,
    ):
        measurement = Measurement(name=f"{self.measurement_name} {data_type}")

        if data_type == "spectrum":
            frequency = Parameter("frequency", label="f", unit="Hz")
            independent_param = frequency
            # TODO: Adjust unit/Name to settings
            signal = Parameter("signal", label=r"$\sqrt{S}$", unit=r"$V/\sqrt{Hz}$")
            dependent_param = signal

            x = results["f_processed"]
            y = results["S_processed"][0]
        elif data_type == "timetrace":
            timer = ElapsedTimeParameter("time")
            independent_param = timer
            dependent_param = self.dependent_param
            # TODO: n_pts is not correct?
            x = np.arange(len(results["timetrace_raw"][0])) / results["settings"].fs
            y = results["timetrace_raw"][0]
        else:
            raise NameError(f"{data_type} is no valid data_type!")

        measurement.register_parameter(independent_param)
        measurement.register_parameter(dependent_param, setpoints=[independent_param])
        static_gettables = []
        for parameter, channel in zip(self.static_gettable_parameters, self.static_gettable_channels):
            measurement.register_parameter(
                channel,
                setpoints=[
                    independent_param,
                ],
            )
            parameter_value = self.properties[parameter["gate"]][parameter["parameter"]]["value"]
            static_gettables.append((channel, [parameter_value for _ in range(len(x))]))

        with measurement.run() as datasaver:
            datasaver.add_result(
                (independent_param, x),
                (dependent_param, abs(y)),
                *static_gettables,
            )
        # TODO: Check if absolut is ok

    # def save_data_to_db(self,
    #                     results,
    #                     indices: int|list,
    #                     data_type: str = "both"):
    #     if type(indices)==int:
    #         self._save_data_to_db(
    #             results=results[indices]):
    #             data_type="spectrum")



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Daniel Grothe


# pylint: disable=missing-function-docstring
import pytest
from qcodes.instrument import Instrument, VisaInstrument
from qcodes.instrument_drivers.mock_instruments import DummyInstrument
from qcodes.parameters import Parameter

from qumada.instrument.instrument import is_instrument_class


@pytest.mark.parametrize(
    "cls,expected",
    [
        (Instrument, True),
        (DummyInstrument, True),
        (VisaInstrument, True),
        (Parameter, False),
    ],
)
def test_is_instrument(cls, expected: bool):
    assert is_instrument_class(cls) is expected


def test_mfli_driver():
    MFLI = pytest.importorskip("qumada.instrument.custom_drivers.ZI.MFLI")
    assert is_instrument_class(MFLI.MFLI)



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Daniel Grothe

# The idea is to get the current version from versioningit for editable installed qumada.
# For built packages, versioningit's onbuild functionality replaces __version__'s value
# with a fixed string.
#
# See
# - https://github.com/QCoDeS/Qcodes/blob/master/qcodes/_version.py
# - https://github.com/jwodder/versioningit/issues/8


def get_version() -> str:
    from pathlib import Path

    import versioningit

    import qumada

    qumada_path = Path(qumada.__file__).parent
    return versioningit.get_version(project_dir=qumada_path.parent.parent)


__version__ = "2.0.0"



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Till Huckeman


import sys
from subprocess import Popen

from qcodes.monitor.monitor import Monitor
from qcodes.station import Station

from qumada.instrument.mapping.base import filter_flatten_parameters
from qumada.measurement.measurement import MeasurementScript


def open_web_gui(parameters):
    """
    Opens Gui from qcodes.monitor.monitor
    parameters: Provides the parameters to display. Has to be Station object,
    QuMADA MeasurementScript object or list of qcodes parameters.
    When a station object is used, all parameters of all components are shown.
    """
    if isinstance(parameters, Station):
        params = [val for val in filter_flatten_parameters(parameters.components).values()]
    elif isinstance(parameters, MeasurementScript):
        params = []
        try:
            channels = [val for val in parameters.gate_parameters.values()]
        except Exception:
            print("Error not yet implemented. Maybe you forgot to do the mapping first?")
            return False
        for gate in channels:
            for item in gate.values():
                params.append(item)
    elif isinstance(parameters, list):
        params = parameters
    else:
        print(
            "The provided parameters are invalid. Please pass as Station \
              object, a Measurement Script (after parameter mapping) or a \
              list of parameters"
        )
        return False
    monitor_process = Popen([sys.executable, "-m", "qcodes.monitor.monitor"], shell=True)
    monitor = Monitor(*params)
    return monitor, monitor_process



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Till Huckeman
# - Daniel Grothe
# - Sionludi Lab


import numpy as np


# %%
def flatten_array(lst) -> list:
    """
    Flattens nested lists and arrays, returns flattened list
    """
    results = []

    def rec(sublist, results):
        for entry in sublist:
            if isinstance(entry, list) or isinstance(entry, np.ndarray):
                rec(entry, results)
            else:
                results.append(entry)

    rec(lst, results)
    return results


# %%


def _validate_mapping(entry, valid_entries, mapping: dict = None, default=None, default_key_error=None):
    """
    Returns mapped value with validation check.

    Parameters
    ----------
    entry : Entry to check.
    valid_entries : Allowed return items.
    mapping : Dictionary mapping the entry to a return value. The default is None (returning entry if valid)
    default : TYPE, optional
        Return value if return value is not valid. The default is None.
    default_key_error : TYPE, optional
        Return value if entry is None or not a key. The default is None.

    Returns
    -------
    Entry if mappiong is None, item mapped to key in mapping if both are valid, else default or default
    key error.

    """
    if not mapping:
        if entry in valid_entries:
            return entry
        else:
            print(f"{entry} is not in {valid_entries}. Using default value: {default}")
            return default
    if entry in mapping.keys():
        if entry in valid_entries:
            return mapping.get(entry)
        else:
            print(f"{mapping.get(entry)} is not in {valid_entries}. Using default value: {default}")
            return default
    else:
        print(f"{entry} is not in mapping. Using default value: {default_key_error}")
        return default_key_error


# %%
def naming_helper(measurement_script, default_name="Measurement"):
    """
    Handles the naming of measurements for measurement scripts.
    Uses the available name with the highest priority.
    Priorities:
        Metadata object in measurement script
        measurement_script.measurement_name
        default_name
    If measurement_script.auto_naming is True, the default name is used always!
    Changes measurement_script.measurement_name to highest priority name
    and returns it.
    If metadata object is available the name in the metadata object is also
    changed!
    """
    if measurement_script.settings.get("auto_naming", False):
        if measurement_script.metadata is not None:
            measurement_script.metadata.measurement.name = default_name
        measurement_script.measurement_name = default_name
    else:
        if measurement_script.metadata is not None:
            measurement_script.measurement_name = measurement_script.metadata.measurement.name
        elif getattr(measurement_script, "measurement_name", None) is not None:
            pass
        else:
            measurement_script.measurement_name = default_name
    return measurement_script.measurement_name



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Daniel Grothe
# - Sionludi Lab


import copy

import numpy as np


def generate_sweep(start, stop, num_points, backsweep=False):
    """
    Creates a linear array with equally spaced setpoints ranging from start to stop.
    If include_backwars is set True, the array will contain twice as much
    setpoints going from start to stop and back down to start again.
    """
    if backsweep:
        sweep = np.array((*np.linspace(start, stop, num_points), *np.linspace(stop, start, num_points)))
    else:
        sweep = np.linspace(start, stop, num_points)
    return sweep


def replace_parameter_settings(parameters: dict, old_val: str, new_value):
    """
    Replaces parameters based on their values with other value. Can be used to
    pass setpoint arrays to parameter-dicts created from json files without
    having to change the values by hand everytime.
    """
    for gate, param in parameters.items():
        for x, y in param.items():
            for name, val in y.items():
                if val == old_val:
                    parameters[gate][x][name] = new_value
    return parameters


def update_parameter_settings(parameters: dict, old_val: str, new_value):
    """
    Replaces parameters based on their values with other value. Can be used to
    pass setpoint arrays to parameter-dicts created from json files without
    having to change the values by hand everytime.
    Does not modify the original dict, but return a modified deep copy.
    """
    updated_parameters = copy.deepcopy(parameters)
    for gate, param in updated_parameters.items():
        for x, y in param.items():
            for name, val in y.items():
                if val == old_val:
                    updated_parameters[gate][x][name] = new_value
    return updated_parameters


def parse_code_from_json(parameters):
    """executes simple chunks of python code in strings starting with _"""
    updated_parameters = copy.deepcopy(parameters)
    for gate, param in updated_parameters.items():
        for x, y in param.items():
            for name, val in y.items():
                if isinstance(val, str):
                    if val[0] == "_":
                        updated_parameters[gate][x][name] = eval(val[1:])
                        print(val)
    return updated_parameters



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Daniel Grothe


from __future__ import annotations

from qumada.measurement.measurement import MeasurementScript
from qumada.metadata import BasicMetadata, Metadata


class Job:
    def __init__(
        self,
        script: MeasurementScript,
        metadata: Metadata | None = None,
        parameters: dict | None = None,
    ):
        self._script: MeasurementScript = script
        self._metadata: Metadata = metadata or BasicMetadata()
        self._parameters: dict = parameters or {}



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - 4K User
# - Daniel Grothe
# - Sionludi Lab
# - Till Huckeman
# - Tobias Hangleiter


"""
MFLI driver for zhinst-toolkit version 0.3.3 (and above)
"""

import numpy as np
from qcodes.instrument import Instrument

# import QuMADA as qt
from zhinst.toolkit import Session


class MFLI(Instrument):
    """
    Driver for the Zurich-Instruments lockin based on the zhinst.toolkit driver.
    Adds proper parameters with get and set commands to make the MFLI usable
    with QuMADA.
    Not finished, most parameters are still missing.
    """

    def __init__(
        self, name: str, device: str, serverhost: str = "localhost", existing_session: Session = None, **kwargs
    ):
        super().__init__(name, **kwargs)
        if isinstance(existing_session, Session):
            session = existing_session
        else:
            self.session = session = Session(serverhost)
        self.instr = instr = session.connect_device(device)
        if instr._device_type != "MFLI":
            raise TypeError("The type of instrument you are trying to connect is not supported.")

        # self.daq = tk_mfli.DAQModule(self.instr)
        demod0 = self.instr.demods[0]

        self.add_parameter(
            "voltage",
            label="V",
            unit="V",
            get_cmd=lambda: np.sqrt(demod0.sample()["x"] ** 2 + demod0.sample()["y"] ** 2),
            get_parser=float,
            set_cmd=False,
            docstring="Absolute voltage as measured by demod0",
        )

        self.voltage.signal_name = ("demod0", "r")

        self.add_parameter(
            "voltage_offset",
            label="Voltage Offset",
            unit="V",
            get_cmd=lambda: self.instr.sigouts[0].offset(),
            get_parser=float,
            set_cmd=lambda x: self.instr.sigouts[0].offset(x),
            docstring="Offset to voltage output in V",
        )

        self.add_parameter(
            "voltage_y_component",
            label="Y",
            unit="V",
            get_cmd=lambda: demod0.sample()["y"],
            get_parser=float,
            set_cmd=False,
            docstring="X component of sample measured by demod1",
        )
        self.voltage_y_component.signal_name = ("demod0", "y")

        self.add_parameter(
            "voltage_x_component",
            label="X",
            unit="V",
            get_cmd=lambda: demod0.sample()["y"],
            get_parser=float,
            set_cmd=False,
            docstring="X component of sample measured by demod1",
        )
        self.voltage_x_component.signal_name = ("demod0", "x")

        self.add_parameter(
            "current",
            label="R",
            unit="A",
            get_cmd=lambda: np.sqrt(demod0.sample()["x"] ** 2 + demod0.sample()["y"] ** 2),
            get_parser=float,
            set_cmd=False,
            docstring="Absolute current as measured by demod0",
        )
        self.current.signal_name = ("demod0", "r")
        self.add_parameter(
            "current_x_component",
            label="X",
            unit="A",
            get_cmd=lambda: demod0.sample()["x"],
            get_parser=float,
            set_cmd=False,
            docstring="X component of sample measured by demod1",
        )
        self.current_x_component.signal_name = ("demod0", "x")

        self.add_parameter(
            name="phase",
            label="Phase",
            unit="rad",
            get_cmd=lambda: demod0.sample()["phase"],
            get_parser=float,
            set_cmd=False,
            docstring="Phase of the measured current in radians",
        )
        self.phase.signal_name = ("demod0", "phase")
        self.add_parameter(
            "current_y_component",
            label="Y",
            unit="A",
            get_cmd=lambda: demod0.sample()["y"],
            get_parser=float,
            set_cmd=False,
            docstring="X component of sample measured by demod1",
        )
        self.current_y_component.signal_name = ("demod0", "y")

        self.add_parameter(
            name="frequency",
            label="Frequency",
            unit="Hz",
            get_cmd=lambda: demod0.sample()["frequency"],
            get_parser=float,
            set_cmd=lambda f: self.instr.oscs[0].freq(f),
            docstring="Frequency of the oscillator",
        )

        self.add_parameter(
            name="amplitude",
            label="Amplitude",
            unit="V",
            get_cmd=lambda: self.instr.sigouts[0].amplitudes[1](),
            get_parser=float,
            set_cmd=lambda a: self.instr.sigouts[0].amplitudes[1](a),
            docstring="Amplitude of the voltage output",
        )
        # self.amplitude.signal_name = ("demod0", "y")

        # self.add_parameter(
        #     name = "output_enabled",
        #     label = "Output Enabled",
        #     get_cmd = lambda: self.instr.sigouts[0].on,
        #     get_parser = lambda x: int(bool(x)),
        #     set_cmd = lambda x: self.instr.sigouts[0].on(x),
        #     set_parser = int,
        #     get_parser = bool,
        #     docstring = "Turns Output1 on or off"
        #     )

        self.add_parameter(
            name="output_range",
            label="Output range",
            unit="V",
            get_cmd=lambda: self.instr.sigouts[0].range(),
            set_cmd=lambda x: self.instr.sigouts[0].range(x),
            docstring="Range of the output",
        )

        self.add_parameter(
            name="output_enabled",
            label="Output Enabled",
            get_cmd=lambda: self.instr.sigouts[0].on(),
            set_cmd=lambda x: self.instr.sigouts[0].on(x),
            docstring="Turns Output1 on or off",
        )

        self.add_parameter(
            name="demod0_time_constant",
            label="Time constant",
            unit="s",
            get_cmd=lambda: demod0.timeconstant(),
            get_parser=float,
            set_cmd=lambda t: self.instr.demods[0].timeconstant(t),
            docstring="Time constant of the low-pass filter of demod0",
        )

        self.add_parameter(
            name="demod0_order",
            label="Demod0 filter order",
            get_cmd=lambda: demod0.order(),
            set_cmd=lambda x: demod0.order(x),
            docstring="Gets/Sets the order of the demod 0 filter.",
        )

        self.add_parameter(
            name="demod0_aux_in_1",
            label="Demod0 AuxIn 1",
            get_cmd=lambda: demod0.sample()["auxin0"],
            set_cmd=False,
            get_parser=float,
            docstring="Aux In 1 of demod0",
        )
        self.demod0_aux_in_1.signal_name = ("demod0", "auxin0")

        self.add_parameter(
            name="demod0_aux_in_2",
            label="Demod0 AuxIn 2",
            get_cmd=lambda: demod0.sample()["auxin1"],
            set_cmd=False,
            get_parser=float,
            docstring="Aux In 2 of demod0",
        )
        self.demod0_aux_in_2.signal_name = ("demod0", "auxin1")
        self.add_parameter(
            name="demod0_trig_in",
            label="Demod0 Trigger Input",
            get_cmd=None,
            set_cmd=None,
            docstring="Gets/Sets the order of the demod 0 filter.",
        )

    def get_idn(self):
        # TODO: Implement this function!
        # /dev..../features/devtype
        # /dev..../features/serial
        # /dev..../features/options
        return {}
